class FeaMaterial;
class Fea1DStressStrainCM;
class Fea1DMomRotCM;
class Fea2DStressStrainCM;
class FeaGraphicDevice;
class FeaStructure;
/****************************/
/*      FeaElement          */
/****************************/

class FeaAugmentedMatix;
class FeaTransformation;
class FeaElement {
	char* name;
public:
	FeaElement(const char*);
	virtual ~FeaElement();
	void operator=(const FeaElement&);
	FeaElement(const FeaElement&);

	int operator==(const char*);
	virtual FeaTransformation& getConnectingDof() = 0;
	virtual void reattach(FeaStructure& newModel) = 0;

	virtual FeaAugmentedMatix& getStiff() = 0;
	virtual FeaAugmentedMatix& getMass() = 0;
	virtual FeaAugmentedMatix& getDamp() = 0;
	virtual FeaAugmentedVector& getResistingForce()= 0;

	virtual double getEventFactor() = 0;
	virtual void commitState() = 0;
	virtual void updateState() = 0;

	// add by hungming, last modified 5/7/01
	virtual double getEventFactor(list<double>* factlist) = 0;

	virtual FeaElement& copy() = 0;	// make a copy of the element

	virtual void moveBy(FeaGeoVector& vTranslation, FeaFullMatrix& T){;}; // move element by the give translation and rotation (not likely used by anything other than superelements
	virtual void rotateDofBy(FeaFullMatrix& T) = 0; // rotate the dof coord systems by the given rotation matrix

	char* getName(); // return the name of the element
	void appendNameWith(char*); // add to given string to the end of the name

	virtual int isLinearElastic() { return FALSE; };	// by default, all elements are linear-elastic


	virtual void printDebug();
	virtual void printEcho();
	virtual void printForce();
	virtual void plotDeformedShape(FeaGraphicDevice& openDevice, double magFactor){};
	virtual void saveState(fstream& theFile, double theTime){};
	void printName() {cout << name;};
	virtual void printStateHeader(){};
	virtual void printStateInfo(fstream& theFile, double theTime){};
};

/****************************/
/*     FeaElementLoad       */
/****************************/

class FeaAugmentedVector;
class FeaElementLoad : public FeaLoad{
public:
	FeaElementLoad(FeaTimeFunction& ntf):FeaLoad(ntf){};
	FeaElementLoad():FeaLoad(){};
	FeaElementLoad(const FeaElementLoad& other):FeaLoad(other){};
	virtual void operator=(const FeaElementLoad& other){FeaLoad::operator=(other);};

	virtual FeaAugmentedVector& getLoad(double)= 0;
	virtual FeaAugmentedVector& getLoad()= 0;
	virtual FeaAugmentedVector& getDeltaLoad(double)= 0;

	virtual FeaLoad& copy() = 0;
   virtual void reattach(FeaStructure& newModel) = 0;		// reattach the load case to the nodes in the new model

	virtual void printEcho(){};
	virtual void print(){};
};

class FeaNode;
class FeaTrussBarPointLoad;
class FeaTrussBarTemp;
class FeaTrussBarPreStress;
/****************************/
/*       FeaTrussBar        */
/****************************/

class FeaTrussBar : public FeaElement {
protected:
	FeaNode* nodeI; 			// connecting node I
	FeaNode* nodeJ; 			// connecting node J

	Fea1DStressStrainCM* mat; // constitutive model for the bar

	Fea1DRect* cs; 			// internal coordinate system along axis of element
	Fea3DRect* mcs; 			// internal coordinate system for the mass matrix

	double coefOfExp;  		// coefficient of thermal expansion
	double area;    			// Xsect area
	double l;     				// length of bar
	double massPerLength;  	// unit mass of the bar
	double dampPerLength;  	// unit damping of the bar
	double ejlI;  				// I end equivalent joint load
	double ejlJ;  				// J end equivalent joint load
	double internalForce; 	// +ve Tension
public:
	FeaTrussBar(const char*, double, FeaNode&, FeaNode&, Fea1DStressStrainCM&, FeaMaterial&, double);
	FeaTrussBar(FeaTrussBar&);
	void operator=(FeaTrussBar&);
	~FeaTrussBar();

	void updateState();
	void commitState();
	virtual FeaAugmentedVector& getResistingForce();

	virtual FeaAugmentedMatix& getStiff();
	FeaAugmentedMatix& getMass();
	FeaAugmentedMatix& getDamp();
	FeaTransformation& getConnectingDof();
	void reattach(FeaStructure& newModel);

	FeaAugmentedVector& pointLoad(FeaTrussBarPointLoad*);
	FeaAugmentedVector& deltaPointLoad(FeaTrussBarPointLoad*);
	void temperature(FeaTrussBarTemp*);
	void preStress(FeaTrussBarPreStress*);

	double getEventFactor();

	// add by hungming, last modified 8/3/01
	double getEventFactor(list<double>* factlist){ return 0;};
	//double getEventFactor(list<double>* factlist);

	virtual void rotateDofBy(FeaFullMatrix& T); // rotate the dof coord systems by the given rotation matrix

	virtual FeaElement& copy();

	void printDebug();
	void printEcho();
	void printForce();
	void plotDeformedShape(FeaGraphicDevice& openDevice, double magFactor);
	virtual void saveState(fstream& theFile, double theTime);
	virtual void printStateHeader();
	virtual void printStateInfo(fstream& theFile, double theTime);
};

/****************************/
/*   FeaTrussBarPointLoad   */
/****************************/

class FeaAugmentedVector;
class FeaTrussBarPointLoad : public FeaElementLoad {
	FeaTrussBar* el;
	double basicLoad;
	double currentLoad;
	double lastLoad;
	double location;
public:
	FeaTrussBarPointLoad(FeaTrussBar& el, double load, double location, FeaTimeFunction& tf);
	FeaTrussBarPointLoad(FeaTrussBar& el, double load, double location);
	~FeaTrussBarPointLoad(){};
	virtual void operator=(const FeaTrussBarPointLoad& other);
	FeaTrussBarPointLoad(const FeaTrussBarPointLoad& other);

	FeaAugmentedVector& getLoad(double time);
	FeaAugmentedVector& getLoad();
	FeaAugmentedVector& getDeltaLoad(double time);

	double getPointLoad(){return (currentLoad);};
	double getDeltaPointLoad(){return (currentLoad - lastLoad);};
	double getPointLocation(){return (location);};

   virtual FeaLoad& copy();
   virtual void reattach(FeaStructure& newModel);							// reattach the load case to the nodes in the new model

	void print();
	void printEcho();
};

/****************************/
/*  FeaTrussBarTemp         */
/****************************/

class FeaTrussBarTemp : public FeaInitialElementState {
	FeaTrussBar* el;
	double temperature;
public:
	FeaTrussBarTemp(FeaTrussBar& element, double temperatureChange);
	~FeaTrussBarTemp(){};
	void operator=(const FeaTrussBarTemp& other);
	FeaTrussBarTemp(const FeaTrussBarTemp& other);

   virtual FeaInitialElementState& copy();
   virtual void reattach(FeaStructure& newModel);							// reattach the load case to the nodes in the new model

	void apply(){el->temperature(this);};
	double getTemp(){return temperature;};

   void printEcho();
};

/****************************/
/*  FeaTrussBarPreStress    */
/****************************/

class FeaTrussBarPreStress : public FeaInitialElementState {
	FeaTrussBar* el;
	double barForce;
public:
	FeaTrussBarPreStress(FeaTrussBar& el, double tensileForce);
	~FeaTrussBarPreStress(){};
	virtual void operator=(const FeaTrussBarPreStress& other);
	FeaTrussBarPreStress(const FeaTrussBarPreStress& other);

   virtual FeaInitialElementState& copy();
   virtual void reattach(FeaStructure& newModel);							// reattach the load case to the nodes in the new model

	void apply(){el->preStress(this);};
	double getBarForce() { return barForce; };

   void printEcho();
};

/****************************/
/*     FeaTrussBarPDRF      */
/****************************/

class FeaTrussBarPDRF : public FeaTrussBar {
// a truss bar with P-delta correction for the resisting force calculation
	Fea1DRect* csPDRF;	// coordinate system along axis of element in deformed configuration
public:
	FeaTrussBarPDRF(const char*, double, FeaNode&, FeaNode&, Fea1DStressStrainCM&, FeaMaterial&, double);
   	// same constructor as for a regular truss bar
	FeaTrussBarPDRF(FeaTrussBarPDRF& other);
	void operator=(FeaTrussBarPDRF& other);
	~FeaTrussBarPDRF();

	virtual void rotateDofBy(FeaFullMatrix& T); // rotate the dof coord systems by the given rotation matrix
	FeaAugmentedVector& getResistingForce();	// places load in the deformed configuaration axis
   virtual FeaElement& copy();
};

/****************************/
/*     FeaTrussBarPDRFK     */
/****************************/

class FeaTrussBarPDRFK : public FeaTrussBarPDRF {
// a truss bar with P-delta correction for the resisting force calculation and stiffness matrix
	Fea3DRect* csPDRFK;	// 3D coordinate system along axis of element (for stiffness matrix)
public:
	FeaTrussBarPDRFK(const char*, double, FeaNode&, FeaNode&, Fea1DStressStrainCM&, FeaMaterial&, double);
   	// same constructor as for a regular truss bar
	FeaTrussBarPDRFK(FeaTrussBarPDRFK& other);
	void operator=(FeaTrussBarPDRFK& other);
	~FeaTrussBarPDRFK();

	virtual void rotateDofBy(FeaFullMatrix& T); // rotate the dof coord systems by the given rotation matrix

	FeaAugmentedMatix& getStiff();
   virtual FeaElement& copy();
};

/****************************/
/* FeaHingedBeamColumn      */
/****************************/

class FeaTransformation;
class FeaHingedBeamColumn : public FeaElement {
protected:
	FeaNode* nodeI; 				// connecting node I
	FeaNode* nodeJ; 				// connecting node J

	Fea1DMomRotCM* hingeI; 			// material
	Fea1DMomRotCM* hingeJ; 			// material
	Fea1DStressStrainCM* axialMat;	// material

	Fea3DRect* mtcs; 				// 3D translational coordinate system for mass
	Fea2DRect* tcs; 				// translational coordinate system
	Fea1DRect* rcs; 				// rotational coordinate system

	double t00, t01, t10, t11;

	// add by hungming, last modified 5/7/01
	double t00commit, t01commit, t10commit, t11commit;

 	double area;    				// Xsect area
	double EA;    					// Xsect area
	double EI;    					// moment of inertia
	double l;     					// length of elastic bar (between rigid links)
	double ll;						// length of left hand rigid link
	double lr;						// length of right hand rigid link
	double massPerLength; 			// mass per unit length
	double ejlI;  					// I end equivalent joint FeaLoad
	double ejlJ;  					// J end equivalent joint FeaLoad
	double internalForceI; 			// +ve Tension
	double internalForceJ; 			// +ve Tension
	double axialForce;
	double q1Commit;
	double q2Commit;
	double q3Commit;
	double q4Commit;
	double q5Commit;
	double q1Update;
	double q2Update;
	double q3Update;
	double q4Update;
	double q5Update;

	void updateStateE2E();
	void setT();

	// add by hungming, last modified 5/7/01
	void commitT();
public:
	FeaHingedBeamColumn(const char*, double A, double I,
		Fea1DMomRotCM&, Fea1DMomRotCM&, FeaNode&,
		FeaNode&, FeaGeoPoint&,  FeaMaterial&, double lLeft=0.0L, double lRight=0.0L);  // used in a 3D context
	FeaHingedBeamColumn(const char*, double A, double I,
		Fea1DMomRotCM&, Fea1DMomRotCM&, FeaNode&,
		FeaNode&, FeaMaterial&, double lLeft=0.0L, double lRight=0.0L);  // used in a 2D context
	FeaHingedBeamColumn(FeaHingedBeamColumn&);
	void operator=(FeaHingedBeamColumn&);
	virtual ~FeaHingedBeamColumn();

	void updateState();
	void commitState();
	virtual FeaAugmentedVector& getResistingForce();

	virtual FeaAugmentedMatix& getStiff();
	FeaAugmentedMatix& getMass();
	FeaAugmentedMatix& getDamp();
	FeaTransformation& getConnectingDof();
	void reattach(FeaStructure& newModel);

	double getEventFactor();

	// add by hungming, last modified 5/7/01
	double getEventFactor(list<double>* factlist);


	virtual FeaElement& copy();

	virtual void rotateDofBy(FeaFullMatrix& T); // rotate the dof coord systems by the given rotation matrix

	void printDebug();
	void printEcho();
	void printForce();
	virtual void saveState(fstream& theFile, double theTime);
	virtual void printStateHeader();
	virtual void printStateInfo(fstream& theFile, double theTime);
	void plotDeformedShape(FeaGraphicDevice& openDevice, double magFactor);
};


/****************************/
/* FeaHingedBeamColumnPDRF  */
/****************************/

class FeaHingedBeamColumnPDRF : public FeaHingedBeamColumn {
// beam-column with hinges at the ends;  includes P-delta calc in resisting force
	Fea2DRect* tcsPDRF; 	 		// translational coordinate system for resisting force transformation
protected:
	FeaGeoPoint* thirdPoint; 	// additional point used to orient roll of the beam
public:
	FeaHingedBeamColumnPDRF(const char*, double A, double I,
		Fea1DMomRotCM&, Fea1DMomRotCM&, FeaNode&,
		FeaNode&, FeaGeoPoint&,  FeaMaterial&, double lLeft=0.0L, double lRight=0.0L);  // used in a 3D context
	FeaHingedBeamColumnPDRF(const char*, double A, double I,
		Fea1DMomRotCM&, Fea1DMomRotCM&, FeaNode&,
		FeaNode&, FeaMaterial&, double lLeft=0.0L, double lRight=0.0L);  // used in a 2D context
	FeaHingedBeamColumnPDRF(FeaHingedBeamColumnPDRF&);
	void operator=(FeaHingedBeamColumnPDRF&);
	~FeaHingedBeamColumnPDRF();

	virtual void rotateDofBy(FeaFullMatrix& T); // rotate the dof coord systems by the given rotation matrix

	FeaAugmentedVector& getResistingForce();
   virtual FeaElement& copy();
};


/****************************/
/* FeaHingedBeamColumnPDRFK */
/****************************/

class FeaHingedBeamColumnPDRFK : public FeaHingedBeamColumnPDRF {
// beam-column with hinges at the ends;  includes P-delta calc in resisting force and stiffness matrix
	Fea3DRect* tcsPDRFK; 	 		// translational coordinate system for stiffness matrix transformation
public:
	FeaHingedBeamColumnPDRFK(const char*, double A, double I,
		Fea1DMomRotCM&, Fea1DMomRotCM&, FeaNode&,
		FeaNode&, FeaGeoPoint&,  FeaMaterial&, double lLeft=0.0L, double lRight=0.0L);  // used in a 3D context
	FeaHingedBeamColumnPDRFK(const char*, double A, double I,
		Fea1DMomRotCM&, Fea1DMomRotCM&, FeaNode&,
		FeaNode&, FeaMaterial&, double lLeft=0.0L, double lRight=0.0L);  // used in a 2D context
	FeaHingedBeamColumnPDRFK(FeaHingedBeamColumnPDRFK&);
	void operator=(FeaHingedBeamColumnPDRFK&);
	~FeaHingedBeamColumnPDRFK();

	virtual void rotateDofBy(FeaFullMatrix& T); // rotate the dof coord systems by the given rotation matrix

	FeaAugmentedMatix& getStiff();
   virtual FeaElement& copy();
};

/****************************/
/*     FeaPointMass         */
/****************************/

class FeaPointMass : public FeaElement {
	FeaNode* node1; // connecting node
	double mass;
	Fea3DRect* cs;
public:
	FeaPointMass(const char*, double , FeaNode&);
	FeaPointMass(FeaPointMass&);
	void operator=(FeaPointMass&);
	~FeaPointMass();

	void updateState();
	void commitState();
	FeaAugmentedVector& getResistingForce();

	FeaAugmentedMatix& getStiff();
	FeaAugmentedMatix& getMass();
	FeaAugmentedMatix& getDamp();
	FeaTransformation& getConnectingDof();
	void reattach(FeaStructure& newModel);

	double getEventFactor();

	// add by hungming, last modified 5/7/01
	double getEventFactor(list<double>* factlist){ return 0;};


   virtual FeaElement& copy();

	virtual void rotateDofBy(FeaFullMatrix& T); // rotate the dof coord systems by the given rotation matrix

	void printDebug();
	void printEcho();
	void printForce();
};

/****************************/
/*     FeaDKTPlate          */
/****************************/

class FeaDKTPlate : public FeaElement {
	FeaNode* node1; // connecting node
	FeaNode* node2; // connecting node
	FeaNode* node3; // connecting node

   Fea1DRect* tcs;
	Fea2DRect* rcs;

   FeaSymmetricMatrix* k;

   FeaFullVector* nodeReactions;
   FeaFullVector* stress;

   double t;
   double e;
   double poi;
   double xC, yC, zC;
public:
	FeaDKTPlate(const char* elName, FeaNode& n1, FeaNode& n2, FeaNode& n3, 
		double thickness, double youngsMod, double poissonRatio);
	FeaDKTPlate(FeaDKTPlate&);
	void operator=(FeaDKTPlate&);
	~FeaDKTPlate();

	void updateState();
	void commitState();
	FeaAugmentedVector& getResistingForce();

	FeaAugmentedMatix& getStiff();
	FeaAugmentedMatix& getMass();
	FeaAugmentedMatix& getDamp();
	FeaTransformation& getConnectingDof();
	void reattach(FeaStructure& newModel);

	double getEventFactor();

	// add by hungming, last modified 5/7/01
	double getEventFactor(list<double>* factlist){ return 0;};


	virtual FeaElement& copy();

	virtual void rotateDofBy(FeaFullMatrix& T); // rotate the dof coord systems by the given rotation matrix

	void printDebug();
	void printEcho();
	void printForce();
	void plotDeformedShape(FeaGraphicDevice& openDevice, double magFactor);
	virtual void saveState(fstream& theFile, double theTime);
	virtual void printStateHeader();
	virtual void printStateInfo(fstream& theFile, double theTime);
};

/****************************/
/*    FeaElasticBeam        */
/****************************/
class FeaElasticBeam : public FeaElement {
protected:
	FeaNode* nodeI; 				// connecting node I
	FeaNode* nodeJ; 				// connecting node J

	Fea3DRect* tcs; 				// translational coordinate system
	Fea3DRect* rcs; 				// rotational coordinate system

	double EA;    					// Xsect area
	double EI12;    				// moment of inertia Strong axis
	double EI34;    				// moment of inertia Weak axis
	double l;     					// length of bar
   double GJ;						// torsional stiffness
   double s1;						// strong bending node i
   double s2;						// strong bending node j
   double s3;						// weak bending node i
   double s4;						// weak bending node j
   double s5;						// axial force
   double s6;						// torque
public:
	FeaElasticBeam(const char* name, double A, double J, double IStrong,
   	double IWeak, FeaNode& ni,FeaNode& nj, FeaGeoPoint& yAxisPoint,  FeaMaterial& mat);
	FeaElasticBeam(FeaElasticBeam&);
	void operator=(FeaElasticBeam&);
	virtual ~FeaElasticBeam();

	void updateState();
	void commitState();
	virtual FeaAugmentedVector& getResistingForce();

	virtual FeaAugmentedMatix& getStiff();
	FeaAugmentedMatix& getMass();
	FeaAugmentedMatix& getDamp();
	FeaTransformation& getConnectingDof();
	void reattach(FeaStructure& newModel);

	double getEventFactor();

	// add by hungming, last modified 5/7/01
	double getEventFactor(list<double>* factlist){ return -1;};


   virtual FeaElement& copy();

	virtual void rotateDofBy(FeaFullMatrix& T); // rotate the dof coord systems by the given rotation matrix

	void printDebug();
	void printEcho();
	void printForce();
	void plotDeformedShape(FeaGraphicDevice& openDevice, double magFactor);
	virtual void saveState(fstream& theFile, double theTime);
	virtual void printStateHeader();
	virtual void printStateInfo(fstream& theFile, double theTime);
};

/****************************/
/*        FeaCST            */
/****************************/

class FeaCST : public FeaElement {
	FeaNode* node1; // connecting node
	FeaNode* node2; // connecting node
	FeaNode* node3; // connecting node

	Fea2DRect* tcs;
	//FeaFullVector* stress;
	double t;
	double area;
	Fea2DStressStrainCM* cm;

	FeaFullMatrix& getB();	// return the b matrix
public:
	FeaCST(const char* elName, FeaNode& n1, FeaNode& n2, FeaNode& n3, double thickness, 
		double youngsMod, double poissonRatio);
	FeaCST(const char* elName, FeaNode& n1, FeaNode& n2, FeaNode& n3, double thickness, 
		Fea2DStressStrainCM& ncm);
	FeaCST(FeaCST&);
	void operator=(FeaCST&);
	~FeaCST();

	void updateState();
	void commitState();
	FeaAugmentedVector& getResistingForce();

	FeaAugmentedMatix& getStiff();
	FeaAugmentedMatix& getMass();
	FeaAugmentedMatix& getDamp();
	FeaTransformation& getConnectingDof();
	void reattach(FeaStructure& newModel);

	double getEventFactor();

	// add by hungming, last modified 5/7/01
	double getEventFactor(list<double>* factlist){ return 0;};


   virtual FeaElement& copy();

	virtual void rotateDofBy(FeaFullMatrix& T); // rotate the dof coord systems by the given rotation matrix

	void printDebug();
	void printEcho();
	void printForce();
	void plotDeformedShape(FeaGraphicDevice& openDevice, double magFactor);
	virtual void saveState(fstream& theFile, double theTime);
	virtual void printStateHeader();
	virtual void printStateInfo(fstream& theFile, double theTime);
};

/****************************/
/*       FeaSpring          */
/****************************/

class FeaSpring : public FeaElement {
protected:
	FeaNode* nodeI; 				// connecting node I

	Fea1DStressStrainCM* mat; 	// constitutive model for the bar

	Fea1DRect* cs; 				// internal coordinate system along axis of spring
   
   dofType type;					// type of degree of freedom of the spring
	double area;    				// Xsect area
	double internalForce; 		// +ve Tension, spring is on the tail side of the direction vector
   									// ie. for horizontal beam support, below beam if direction vector is up
public:
	FeaSpring(const char* name, double area, FeaNode& nodeI, Fea1DRect& cs, 
		dofType type, Fea1DStressStrainCM& cm);
	FeaSpring(FeaSpring&);
	void operator=(FeaSpring&);
	~FeaSpring();

	void updateState();
	void commitState();
	virtual FeaAugmentedVector& getResistingForce();

	virtual FeaAugmentedMatix& getStiff();
	FeaAugmentedMatix& getMass();
	FeaAugmentedMatix& getDamp();
	FeaTransformation& getConnectingDof();
	void reattach(FeaStructure& newModel);

	double getEventFactor();

	// add by hungming, last modified 5/7/01
	double getEventFactor(list<double>* factlist){ return -1;};


   virtual FeaElement& copy();

	virtual void rotateDofBy(FeaFullMatrix& T); // rotate the dof coord systems by the given rotation matrix

	void printDebug();
	void printEcho();
	void printForce();
	virtual void saveState(fstream& theFile, double theTime);
	virtual void printStateHeader();
	virtual void printStateInfo(fstream& theFile, double theTime);
};





